#include "is/isrepository.h"

#include <owl/time.h>

using namespace std;

static char * my_strdup( const char * s )
{
    int i = strlen( s );
	
    char * d = new char[i+1];
	
    memcpy( d, s, i+1 );
    return d;
}
    
 //////////////////////////////////////////////////
 // ISRepository methods implementation
 //////////////////////////////////////////////////

ISRepository::ISRepository( const IPCPartition& p, const char * sid, unsigned long max_history_size, char verbosity )
  : IPCNamedObject<POA_is::repository> ( p, sid ),
    verbosity_ ( verbosity ),
    max_history_size_ ( max_history_size )
{
    dictionary_ = new ISDictionary( information_, mutex_, verbosity );
    if ( publish() == false )
    {
	cerr << OWLTime() << " :: ERROR [ISRepository::ISRepository()] Cannot publish \"" 
		  << sid << "\" object in the partition " << p.name() << endl;
	exit(1);
    }
	    
    if ( verbosity_ > 0 )
    {
	cout << OWLTime() << " :: IS server " << sid << " is created in partition " 
		  << p.name() << endl;
    }
}

ISRepository::ISRepository( const char * sid, unsigned long max_history_size, char verbosity )
  : IPCNamedObject<POA_is::repository> ( sid ),
    verbosity_ ( verbosity ),
    max_history_size_ ( max_history_size )
{
    dictionary_ = new ISDictionary( information_, mutex_, verbosity );

    if ( publish() == false )
    {
	cerr  << OWLTime() << " :: ERROR [ISRepository::ISRepository()] Cannot publish \"" 
			   << sid << "\" object in the partition " 
			   << partition().name() << endl;
	exit(1);
    }
	
    if ( verbosity_ > 0 )
    {
	cout << OWLTime() << " :: INFORMATION [ISRepository::ISRepository()] IS server " 
			  << sid << " is created for partition " <<  partition().name() << endl;
    }
}

ISRepository::~ISRepository( )
{
    dictionary_->_destroy();
    
    InfoMap::iterator it = information_.begin( );
    while ( it != information_.end( ) )
    {
	delete (*it).second;
	information_.erase( it++ );
    }
    
    if ( verbosity_ > 0 )
    {
	cout << OWLTime() << " :: INFORMATION [ISRepository::ISRepository()] IS server " 
			  << name() << " is destroyed for partition " << partition().name() << endl;
    }
    
    if ( withdraw() != true )
    {
	cerr << "WARNING [ISRepository::~ISRepository()] can not unpublish ISRepository object" << endl;
    }
}

void ISRepository::create ( const char * name, const char * data )
{
    if ( verbosity_ > 0 )
    {
	cout << OWLTime() << " :: INFORMATION [ISRepository::create] Request to create the "
			  << name << " information object" << endl;
    }
    
    string key = name;
    
    mutex_.lock();
    InfoMap::iterator it = information_.find( key );
    ISInfoTrue * info = ( it != information_.end() ? (*it).second : 0 );

    if ( info != 0 )
    {
	if ( info -> state() != ISInfoTrue::Public )
	{
	    info -> validate( data, verbosity_ );
	}
	else
	{
	    if ( verbosity_ > 0 )
	    {
		cout << OWLTime() << " :: WARNING [ISRepository::create] Information object " 
				  << name << " already exist" << endl;
	    }
	    mutex_.unlock();
	    throw is::AlreadyExist();
	}
    }    
    else
    {
	information_[ key ] = new ISInfoTrue( name, data, max_history_size_ );
    }
    mutex_.unlock();
    
    inform( name, data, is::Created );
    
    if ( verbosity_ > 0 )
    {
	cout << OWLTime() << " :: INFORMATION [ISRepository::create] Information object " 
			  << name << " is created" << endl;
    }
}

void ISRepository::remove ( const char * name )
{
    if ( verbosity_ > 0 )
    {
	cout << OWLTime() << " :: INFORMATION [ISRepository::remove] Request to remove the "
			  << name << " information object" << endl;
    }

    string key = name;
    
    mutex_.lock();
    InfoMap::iterator it = information_.find( key );
    ISInfoTrue * info = ( it != information_.end() ? (*it).second : 0 );
    
    if ( info == 0 || info -> state() != ISInfoTrue::Public )
    {
	if ( verbosity_ > 0 )
	{
	    cout << OWLTime() << " :: WARNING [ISRepository::remove] Information object " 
	    		      << name << " does not exist" << endl;
	}
	mutex_.unlock();
	throw is::NotFound();
    }
    
    string data = info -> data( );
    if ( info -> hasSubscriptions() == true || info -> frozen() )
    {
	info -> invalidate( verbosity_ );
	mutex_.unlock();
	inform( name, data.c_str(), is::Deleted );
    }
    else
    {
	information_.erase( key );
	mutex_.unlock();
	inform( name, data.c_str(), is::Deleted );
	delete info;
    }

    if ( verbosity_ > 0 )
    {
	cout << OWLTime() << " :: INFORMATION [ISRepository::remove] Information object " 
			  << name << " is destroyed" << endl;
    }
}

void    ISRepository::update (  const char * name, const char * data, 
				bool create_if_not_exist, bool keep_history )
{
    if ( verbosity_ > 0 )
    {
	cout << OWLTime() << " :: INFORMATION [ISRepository::update] Request to update the "
			  << name << " information object (keep_history = " << keep_history << ")" << endl;
    }
    
    string key = name;
    
    mutex_.lock();
    InfoMap::iterator it = information_.find( key );
    ISInfoTrue * info = ( it != information_.end() ? (*it).second : 0 );
    
    if ( info != 0 && info -> state() == ISInfoTrue::Public )
    {
	if ( info -> update( data, keep_history, verbosity_ ) == true )
	{
	    mutex_.unlock();
	    inform( name, data, is::Updated );
	    if ( verbosity_ > 0 )
	    {
		cout << OWLTime() << " :: INFORMATION [ISRepository::update] Information object " 
				  << name << " is updated" << endl;
	    }
	}
	else
	{
	    mutex_.unlock();
	    if ( verbosity_ > 0 )
	    {
		cout << OWLTime() << " :: WARNING [ISRepository::update] Information object " 
				  << name << " has a different type" << endl;
	    }
	}
    }
    else if ( create_if_not_exist )
    {
	mutex_.unlock();
	create( name, data );
    }
    else
    {
	mutex_.unlock();
	if ( verbosity_ > 0 )
	{
	    cout << OWLTime() << " :: WARNING [ISRepository::update] Information object " 
	    		      << name << " does not exist" << endl;
	}
	throw is::NotFound();
    }
}

void    ISRepository::find ( const char * name, CORBA::String_out data )
{
    if ( verbosity_ > 0 )
    {
	cout << OWLTime() << " :: INFORMATION [ISRepository::find] Request to find the "
			  << name << " information object" << endl;
    }
    
    string key = name;
    
    mutex_.lock();
    InfoMap::iterator it = information_.find( key );
    ISInfoTrue * info = ( it != information_.end() ? (*it).second : 0 );
    
    if ( info != 0 && info -> state() == ISInfoTrue::Public )
    {
	data = my_strdup(info -> data());
	mutex_.unlock();
	if ( verbosity_ > 0 )
	{
	    cout << OWLTime() << " :: INFORMATION [ISRepository::find] Information object " 
	    		      << name << " is found" << endl;
	}
    }
    else
    {
	mutex_.unlock();
	if ( verbosity_ > 0 )
	{
	    cout << OWLTime() << " :: WARNING [ISRepository::find] Information object " 
	    		      << name << " does not exist" << endl;
	}
	throw is::NotFound();
    }
}

void ISRepository::history( const char * name, long how_many, is::InfoList_out list )
{
    if ( verbosity_ > 0 )
    {
	cout << OWLTime() << " :: INFORMATION [ISRepository::history] Request for the history of the"
			  << name << " information object (how_many = " << how_many << ")" << endl;
    }
    
    string key = name;
    
    mutex_.lock();
    InfoMap::iterator it = information_.find( key );
    ISInfoTrue * info = ( it != information_.end() ? (*it).second : 0 );
    
    list = new is::InfoList( 0 );
    if ( info != 0 && info -> state() == ISInfoTrue::Public )
    {
	const deque<string> & qu = info -> history();
	long length = ( ( how_many > (long)qu.size() || how_many == -1 ) 
				? (long)qu.size() 
				: how_many );
	    
	list -> length( length < 0 ? 0 : length );
	for ( long i = 0; i < length; i++ )
	{
	    list[i] = my_strdup(qu[i].c_str());
	}
	
	mutex_.unlock();
	if ( verbosity_ > 0 )
	{
	    cout << OWLTime() << " :: INFORMATION [ISRepository::history] Information object " 
	    		      << name << " is found" << endl;
	}
    }
    else
    {
	mutex_.unlock();
	if ( verbosity_ > 0 )
	{
	    cout << OWLTime() << " :: WARNING [ISRepository::history] Information object " 
	    		      << name << " does not exist" << endl;
	}
	throw is::NotFound();
    }
}

void ISRepository::list ( const char * rexp, is::InfoList_out list, is::dictionary_out dict )
{
    
    OWLRegexp re( rexp );
    
    if ( re.status() != OWLRegexp::Ok )
    {
	if ( verbosity_ > 0 )
	{
	    cout << OWLTime() << " :: WARNING [ISRepository::list] Invalid regular expression (" 
	    		      << rexp << ")" << endl;
	}
	throw is::InvalidExpression();
    }
    if ( verbosity_ > 0 )
    {
	cout << OWLTime() << " :: INFORMATION [ISRepository::list] request with regular expression (" 
			  << rexp << ")" << endl;
    }
	
    string rexpression = name();
    rexpression += "\\.";
    rexpression += rexp;
    
    re = rexpression;
	    
    InfoMap::iterator it;
        
    OWLMutexLock ml( mutex_ );
    list = new is::InfoList( information_.size() );
    list -> length( information_.size() );
    size_t idx = 0;
    for ( it = information_.begin( ); it != information_.end( ); ++it )
    {
	ISInfoTrue * info = (*it).second;
	if (    info -> state() == ISInfoTrue::Public 
	     && re.exact_match( (*it).first ) > 0 )
	{
	    info -> freeze();
	    list[idx++] = my_strdup(info -> name());
	}
    }
    list -> length( idx );
    
    dict = is::dictionary::_duplicate( dictionary_->_this() );
}

void ISRepository::subscribe ( const char * pattern, bool is_reg_exp, is::callback_ptr cb )
{    
    if ( is_reg_exp )
    {
	OWLRegexp re( pattern );
	if ( re.status() != OWLRegexp::Ok )
	{
	    if ( verbosity_ > 0 )
	    {
		cout << OWLTime() << " :: WARNING [ISRepository::subscribe] Invalid regular expression (" 
				  << pattern << ")" << endl;
	    }
	    throw is::InvalidExpression();
	}
    
	string rexpression = name();
	rexpression += "\\.";
	rexpression += pattern;
    
	if ( verbosity_ > 0 )
	{
	    cout << OWLTime() << " :: INFORMATION [ISRepository::subscribe] request with regular expression (" 
	    		      << pattern << ")" << endl;
	}
    
	OWLMutexWLock ml( cb_mutex_ );
	callbacks_.push_back( new Callback( rexpression, cb ) );
    }
    else
    {
	string full_name = name();
	full_name += ".";
	full_name += pattern;

	if ( verbosity_ > 0 )
	{
	    cout << OWLTime() << " :: INFORMATION [ISRepository::subscribe] request for the " 
	    		      << full_name << " information object" << endl;
	}
    
	OWLMutexLock ml( mutex_ );
	InfoMap::iterator it = information_.find( full_name );
	ISInfoTrue * info = ( it != information_.end() ? (*it).second : 0 );
    
	if ( info == 0 )
	{
	    info = new ISInfoTrue( full_name.c_str(), max_history_size_ );
	    information_[ full_name ] = info;
	}
	info -> subscribe( cb );
    }
}

void ISRepository::unsubscribe ( const char * pattern, bool is_reg_exp, is::callback_ptr cb )
{
	
    if ( is_reg_exp == true )
    {
	if ( verbosity_ > 0 )
	{
	    cout << OWLTime() << " :: INFORMATION [ISRepository::unsubscribe] request with regular expression (" 
	    		      << pattern << ")" << endl;
	}
	
	OWLMutexWLock ml( cb_mutex_ );
	for ( CallbackMap::iterator it = callbacks_.begin(); it != callbacks_.end(); it++ )
	{
	    if ( (*it)->callback_->_is_equivalent( cb ) )
	    {
		if ( verbosity_ > 0 )
		{
		    cout << OWLTime() << " :: INFORMATION [ISRepository::unsubscribe] unsubscribe successfull " << endl;
		}
		delete *it;
		callbacks_.erase( it );
		return;
	    }
	}
	if ( verbosity_ > 0 )
	{
	    cout << OWLTime() << " :: WARNING [ISRepository::unsubscribe] subscription with regular expression (" 
			      << pattern << ") does not exist" << endl;
	}
	throw is::NotFound();
    }
    else
    {
	if ( verbosity_ > 0 )
	{
	    cout << OWLTime() << " :: INFORMATION [ISRepository::unsubscribe] request to object " 
			      << pattern << endl;
	}
	
	string full_name = name();
	full_name += ".";
	full_name += pattern;
	
	OWLMutexLock ml( mutex_ );
	InfoMap::iterator it = information_.find( full_name );
	ISInfoTrue * info = ( it != information_.end() ? (*it).second : 0 );
	
	if ( info == 0 || info -> unsubscribe( cb ) == false )
	{
	    if ( verbosity_ > 0 )
	    {
		cout << OWLTime() << " :: INFORMATION [ISRepository::unsubscribe] subscription to object " 
				  << pattern << " does not exist" << endl;
	    }
	    throw is::NotFound();	    
	}
    
	if (	   info -> hasSubscriptions() == false
		&& info -> state() == ISInfoTrue::Private
	    && !info -> frozen() )
	{
	    information_.erase( full_name );
	    delete info;
	}
    }
}

void ISRepository::inform ( const char * name, const char * data, is::Reason reason )
{    
    if ( callbacks_.empty() )
    {
	return;
    }
	
    bool errors = false;
    cb_mutex_.read_lock();
    CallbackMap::iterator it = callbacks_.begin();
    
    for ( ; it != callbacks_.end(); ++it )
    {
	if ( (*it)->rexpression_.exact_match( name ) <= 0 )
	    continue;
		
	try
	{
	    (*it)->callback_ -> inform( data, reason );
	    if ( verbosity_ > 0 )
	    {
		cout << OWLTime() << " :: INFORMATION [ISRepository::inform] subscriber for the (" 
				  << (*it)->rexpression_.str() << ") has been notified" << endl;
	    }
	}
	catch(CORBA::SystemException & ex)
	{
	    if ( verbosity_ > 0 )
	    {
		cout << OWLTime() << " :: WARNING [ISRepository::inform] got " << ex._name() << " exception" 
				  << " for the " << (*it)->rexpression_.str() << " subscriber" << endl;
	    }
	    errors = true;
	}
    }
    cb_mutex_.unlock();
    
    if ( errors )
    {
	OWLMutexWLock ml( cb_mutex_ );
	CallbackMap::iterator it = callbacks_.begin();
    
	while ( it != callbacks_.end() )
	{
	    try
	    {
		if ( (*it)->callback_ -> _non_existent( ) )
		{
		    if ( verbosity_ > 0 )
		    {
			cout << OWLTime() << " :: INFORMATION [ISRepository::inform] removing subscription for the (" 
				  << (*it)->rexpression_.str() << ")" << endl;
		    }
		    delete *it;
		    it = callbacks_.erase( it );
		    continue;
		}
	    }
	    catch(...)
	    {
		if ( verbosity_ > 0 )
		{
		    cout << OWLTime() << " :: INFORMATION [ISRepository::inform] removing subscription for the (" 
					<< (*it)->rexpression_.str() << ")" << endl;
		}
		delete *it;
		it = callbacks_.erase( it );
		continue;
	    }
	    it++;
	}
    }
}

